unit HtmlUnit;

interface

uses
    SysUtils, Classes, Contnrs, IniFiles;

type
    //  \NX
    THighlight = class
    private
        FText: string;
        FFoundPos: integer;
    protected
        function    GetLength(): integer; virtual; abstract;
        function    Match(const S: string; const Offset: integer;
                          var FoundPos: integer): Boolean; virtual; abstract; 
    public
        constructor Create(const S: string);
        property Len: integer read GetLength;
        property FoundPos: integer read FFoundPos;
    end;

    //  萔NX
    THighlightString = class(THighlight)
    private
        FText: char;
        FLen: integer;
    protected
        function    GetLength(): integer; override;
        function    Match(const S: string; const Offset: integer;
                          var FoundPos: integer): Boolean; override;
    public
        constructor Create(const S: char);
    end;

    //  1sRgNX
    THighlightComment = class(THighlight)
    private
        FLen: integer;
    protected
        function    GetLength(): integer; override;
        function    Match(const S: string; const Offset: integer;
                          var FoundPos: integer): Boolean; override;
    end;

    //  \ꏈNX
    THighlightReserved = class(THighlight)
    protected
        function    GetLength(): integer; override;
        function    Match(const S: string; const Offset: integer;
                          var FoundPos: integer): Boolean; override;
    end;

    TSqlToHtml = class
    private
        FSettingFilePath: string;
        FIniFile: TIniFile;
        FWorkList: TStringList;
        FHighlightList: TObjectList;
        function    ConvertLine(const Line: string): string;
        procedure   LoadHighlightList();
    public
        constructor Create(const SettingFilePath: string);
        destructor  Destroy(); override;
        procedure   Convert(SQL: TStrings; HTML: TStrings);
    end;

var
    SqlToHtml: TSqlToHtml = nil;

implementation

//-----------------------------------------------------------------------------
//  "<"">""&"AHTMLŐ\邽߂̕ϊ
function ChangeHTMLSpecialChar(text: string): string;
var
    i: integer;
    src, dest: array[0..2] of string;
begin

    //ȂƂȂ[vł܂킵Ȃ̂ŔzɁB
    //"&"ŏɕϊƂȂƑςȂƂ
    src[0] := '&';  dest[0] := '&amp;';
    src[1] := '<';  dest[1] := '&lt;';
    src[2] := '>';  dest[2] := '&gt;';

    //ϊRs[
    Result := Text;
    //bAfor i := 0 to 2 do
    for i := Low(src) to High(src) do
    begin
        //uĂ邾
        Result := StringReplace(Result, src[i], dest[i], [rfReplaceAll]);
    end;

end;

//-----------------------------------------------------------------------------
//  AvP[Vf[^tH_
function    GetDataDir(): string;
begin
    Result := ExtractFilePath(ParamStr(0)) + 'dat\';
end;

//*****************************************************************************
//  THighlight

//-----------------------------------------------------------------------------
//  RXgN^
constructor THighlight.Create(const S: string);
begin
    FText := S;
end;

//*****************************************************************************
//  THighlightString

//-----------------------------------------------------------------------------
//  RXgN^
constructor THighlightString.Create(const S: char);
begin
    FText := S;
end;

//-----------------------------------------------------------------------------
//  ̒
function    THighlightString.GetLength(): integer;
begin
    Result := FLen;
end;

//-----------------------------------------------------------------------------
//  S̕񌟍
function    THighlightString.Match(const S: string;   //  ׂ1s
                  const Offset: integer;        //  Jnʒu
                  var FoundPos: integer         //  ʒu
                  ): Boolean;                   //  ?
var
    Target: string;
    i, Index, RealIndex, LastPos: integer;
    c, nextc: char;
begin
    Result := false;

    //  ۂ̌Ώە
    Target := Copy(S, Offset, Length(S));

    //  Ƃ肠KɒT
    Index := AnsiPos(UpperCase(FText), UpperCase(Target));
    //  łȂI
    if (Index = 0) then
    begin
        exit;
    end
    //  ܂
    else
    begin
        //  ۂ̈ʒu
        RealIndex := Offset + Index - 1;
        //  Pcʒu
        LastPos := RealIndex; 
        //  Ă܂傤
        i := RealIndex + 1;
        while (i <= Length(S)) do
        begin
            //  1oCgłȂƑʖ
            if (ByteType(S, i) <> mbSingleByte) then
            begin
                Inc(i);
                continue;
            end;

            c := S[i];

            //  ̏I肾낤?
            if (c <> FText) then
            begin
                Inc(i);
                continue;
            end;

            //  2AƂĂȂ낤˂
            if (i < Length(S)) then
            begin
                nextc := S[i + 1];
                if (nextc = FText) then
                begin
                    Inc(i, 2);
                    continue;
                end;
            end;

            //  ܂ł݂
            LastPos := i;
            break;
        end;

        //  ܂ȂƂ?
        if (RealIndex = LastPos) then LastPos := Length(S);

        //  ǂ
        FLen := LastPos - RealIndex + 1;

        //@̑
        FoundPos := RealIndex;
        FFoundPos := FoundPos;
        Result := true;
    end;
end;

//*****************************************************************************
//  THighlightComment

//-----------------------------------------------------------------------------
//  ̒
function    THighlightComment.GetLength(): integer;
begin
    Result := FLen;
end;

//-----------------------------------------------------------------------------
//  S̕񌟍
function    THighlightComment.Match(const S: string;   //  ׂ1s
                  const Offset: integer;        //  Jnʒu
                  var FoundPos: integer         //  ʒu
                  ): Boolean;                   //  ?
var
    Target: string;
    Index, RealIndex, LastPos: integer;
begin
    Result := false;

    //  ۂ̌Ώە
    Target := Copy(S, Offset, Length(S));

    //  T
    Index := AnsiPos(UpperCase(FText), UpperCase(Target));
    //  łȂI
    if (Index = 0) then
    begin
        exit;
    end
    //  ܂
    else
    begin
        //  ۂ̈ʒu
        RealIndex := Offset + Index - 1;
        //  Ō܂łׂđΏ
        LastPos := Length(S); 
        //  ǂ
        FLen := LastPos - RealIndex + 1;
        //@̑
        FoundPos := RealIndex;
        FFoundPos := FoundPos;
        Result := true;
    end;
end;

//*****************************************************************************
//  THighlightReserved

//-----------------------------------------------------------------------------
//  ̒
function    THighlightReserved.GetLength(): integer;
begin
    Result := Length(FText);
end;

//-----------------------------------------------------------------------------
//  S̕񌟍
function    THighlightReserved.Match(const S: string;   //  ׂ1s
                  const Offset: integer;        //  Jnʒu
                  var FoundPos: integer         //  ʒu
                  ): Boolean;                   //  ?
var
    Target: string;
    Index, RealIndex: integer;
begin
    Result := false;
    //  ۂ̌Ώە
    Target := Copy(S, Offset, Length(S));

    while (Target <> '') do
    begin
        //  Ƃ肠KɒT
        Index := AnsiPos(UpperCase(FText), UpperCase(Target));
        //  łȂI
        if (Index = 0) then
        begin
            break;
        end
        //  ܂
        else
        begin
            //  ۂ̈ʒu
            RealIndex := Offset + Index - 1;
            //  O̕ɕςȂ̂˂?
            if (RealIndex > 1) then
            begin
                //  ȊO̕ʖڂł
                if not (S[RealIndex - 1] in [' ', #9, '(', ')']) then
                begin
                    //  ͈͂ɂ炷
                    Target := Copy(Target, Index + Len, Length(S));
                    continue;
                end;
            end;
            //  ̕ɕςȂ̂˂?
            if (RealIndex + Len - 1 < Length(S)) then
            begin
                //  ȊO̕ʖڂł
                if not (S[RealIndex + Len] in [' ', #9, '(', ')']) then
                begin
                    //  ͈͂ɂ炷
                    Target := Copy(Target, Index + Len, Length(S));
                    continue;
                end;
            end;
            //  ̗p
            FoundPos := RealIndex;
            FFoundPos := FoundPos;
            Result := true;
            break;
        end;
    end;
end;

//*****************************************************************************
//  TSqlToHtml

//-----------------------------------------------------------------------------
//  RXgN^
constructor TSqlToHtml.Create(const SettingFilePath: string);
begin
    //  Ɨp
    FWorkList := TStringList.Create();
    //  ݒt@Cێ
    FSettingFilePath := SettingFilePath;
    //  \ꏈNXXg
    FHighlightList := TObjectList.Create(true);
    //  \NX쐬
    LoadHighlightList();
end;
//-----------------------------------------------------------------------------
//  fXgN^
destructor  TSqlToHtml.Destroy();
begin
    FHighlightList.Free();
    FWorkList.Free();
end;
//-----------------------------------------------------------------------------
//  \ꗗǍ
procedure   TSqlToHtml.LoadHighlightList();
var
    i: integer;
    Path: string;
begin

    //  ------------------------------
    //  ݒt@CǍ
    FIniFile := TIniFile.Create(FSettingFilePath);

    //  ------------------------------
    //  ݒێt@CpX擾
    Path := FIniFile.ReadString(Self.ClassName, 'string', 'string.txt');
    Path := GetDataDir() + Path;
    //  ꗗǍ
    FWorkList.LoadFromFile(Path);
    //  NX쐬
    for i := 0 to FWorkList.Count - 1 do
    begin
        FHighlightList.Add(THighlightString.Create(FWorkList[i][1]));
    end;
    FWorkList.Clear();

    //  ------------------------------
    //  Rgݒێt@CpX擾
    Path := FIniFile.ReadString(Self.ClassName, 'string', 'comment.txt');
    Path := GetDataDir() + Path;
    //  ꗗǍ
    FWorkList.LoadFromFile(Path);
    //  NX쐬
    for i := 0 to FWorkList.Count - 1 do
    begin
        FHighlightList.Add(THighlightComment.Create(FWorkList[i]));
    end;
    FWorkList.Clear();

    //  ------------------------------
    //  \ێt@CpX擾
    Path := FIniFile.ReadString(Self.ClassName, 'reserved', 'reserved.txt');
    Path := GetDataDir() + Path;
    //  ꗗǍ
    FWorkList.LoadFromFile(Path);
    //  NX쐬
    for i := 0 to FWorkList.Count - 1 do
    begin
        FHighlightList.Add(THighlightReserved.Create(FWorkList[i]));
    end;
    FWorkList.Clear();

    //  ------------------------------
    //  ݒt@C
    FIniFile.Free();
end;
//-----------------------------------------------------------------------------
//  ϊ
procedure   TSqlToHtml.Convert(SQL: TStrings; HTML: TStrings);
var
    i: integer;
begin
    HTML.Clear();
    for i := 0 to SQL.Count - 1 do
    begin
        HTML.Add(ConvertLine(SQL[i]));
    end;

    //  擪Ɩ<pre>ŋ
    if (HTML.Count > 0) then
    begin
        HTML.Insert(0, '<pre class="code">');
        HTML.Append('</pre>');
    end;
end;
//-----------------------------------------------------------------------------
//  1sϊ
function    TSqlToHtml.ConvertLine(const Line: string): string;
var
    i: integer;
    H, OkH: THighlight;
    Offset, FoundPos: integer;
    HiLiteText: string;
    aLine: string;
begin
    //  ꕶ͒uĂ(ȂƂłĂ..?)
    aLine := ChangeHTMLSpecialChar(Line);

    Result := '';

    //  Jnʒu
    Offset := 1;
    //  ̗pꂽ\
    OkH := nil;

    //  Jnʒu̒𒴂瓖RI
    while (Offset <= Length(aLine)) do
    begin

        //  e\Ɍ
        for i := 0 to FHighlightList.Count - 1 do
        begin
            H := THighlight(FHighlightList[i]);
            //  
            if (H.Match(aLine, Offset, FoundPos)) then
            begin
                //  O߂?
                if (OkH = nil) then
                begin
                    //  oĂ
                    OkH := H;
                end
                //  ߂ĂȂ
                else
                begin
                    //  O̓z葁ʒuŌ?
                    if (FoundPos < OkH.FoundPos) then
                    begin
                        //  O̗p
                        OkH := H;
                    end;
                end;
            end;
        end;
        //  ̂
        if (OkH <> nil) then
        begin
            //  \Ă݂悤
            HiLiteText := Copy(aLine, OkH.FoundPos, OkH.Len);

            //  ̏ꍇ
            if (OkH is THighlightString) then
            begin
                HiLiteText := '<em class="string">' + HiLiteText + '</em>';
            end
            //  Rg̏ꍇ
            else if (OkH is THighlightComment) then
            begin
                HiLiteText := '<em class="comment">' + HiLiteText + '</em>';
            end
            //  \̏ꍇ
            else if (OkH is THighlightReserved) then
            begin
                HiLiteText := '<strong class="reserved">' + HiLiteText + '</strong>';
            end
            //  ˂
            else
            begin
                HiLiteText := '<em class="default">' + HiLiteText + '</em>';
            end;

            Result := Result
            //  Jnʒu甭ʒu܂ł͖{
                    + Copy(aLine, Offset, OkH.FoundPos - Offset)
            //  ʒu甭܂ł͋\
                    + HiLiteText;
            //  ̗\͔\ŒT
            Offset := OkH.FoundPos + OkH.Len;
            OkH := nil;
        end
        //  Ȃ
        else
        begin
            Result := Result
                    + Copy(aLine, Offset, Length(aLine));
            //  ̍s͒אs
            break;
        end;
    end;
end;

//-----------------------------------------------------------------------------
initialization
    SqlToHtml := TSqlToHtml.Create(ExtractFilePath(ChangeFileExt(ParamStr(0), '.ini')));
//-----------------------------------------------------------------------------
finalization
    SqlToHtml.Free();
//-----------------------------------------------------------------------------
end.

